@@ -5,7 +5,7 @@ use Carp qw/croak confess carp/;
 
 BEGIN {
   $Math::Prime::Util::GMP::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::GMP::VERSION = '0.41';
+  $Math::Prime::Util::GMP::VERSION = '0.50';
 }
 
 # parent is cleaner, and in the Perl 5.10.1 / 5.12.0 core, but not earlier.
@@ -62,14 +62,15 @@ our @EXPORT_OK = qw(
                      sigma
                      chinese
                      moebius
-                     prime_count
+                     prime_count prime_count_lower prime_count_upper
                      primorial
                      pn_primorial
-                     factorial
+                     factorial factorialmod
                      consecutive_integer_lcm
                      partitions bernfrac bernreal harmfrac harmreal stirling
-                     zeta riemannr
-                     gcd lcm kronecker valuation binomial gcdext
+                     zeta li ei riemannr lambertw
+                     logreal expreal powreal agmreal
+                     gcd lcm kronecker valuation binomial gcdext hammingweight
                      invmod sqrtmod addmod mulmod divmod powmod
                      vecsum vecprod
                      exp_mangoldt
@@ -77,33 +78,62 @@ our @EXPORT_OK = qw(
                      totient
                      jordan_totient
                      carmichael_lambda
-                     sqrtint rootint
-                     is_power is_prime_power
+                     sqrtint rootint logint
+                     is_power is_prime_power is_semiprime is_square
+                     is_carmichael is_fundamental is_totient
                      is_primitive_root
+                     is_polygonal polygonal_nth
                      znorder
                      znprimroot
                      ramanujan_tau
-                     Pi
+                     Pi Euler
                      todigits
+                     random_prime random_nbit_prime random_ndigit_prime
+                     random_strong_prime
+                     random_maurer_prime random_shawe_taylor_prime
+                     random_maurer_prime_with_cert
+                     random_shawe_taylor_prime_with_cert
+                     seed_csprng is_csprng_well_seeded
+                     irand irand64 drand urandomb urandomm urandomr random_bytes
+                     permtonum numtoperm
                    );
                    # Should add:
                    # nth_prime
 our %EXPORT_TAGS = (all => [ @EXPORT_OK ]);
 
+sub _init_random {
+  use Fcntl;
+  my($file, $nbytes, $s, $buffer, $nread) = ("/dev/urandom", 256, '', '', 0);
+  return if $^O eq 'MSWin32';
+  return unless -r $file;
+  sysopen(my $fh, $file, O_RDONLY);
+  binmode $fh;
+  while ($nread < $nbytes) {
+    my $thisread = sysread $fh, $buffer, $nbytes-$nread;
+    last unless defined $thisread && $thisread > 0;
+    $s .= $buffer;
+    $nread += length($buffer);
+  }
+  return unless $nbytes == length($s);
+  seed_csprng($nbytes, $s);
+}
+
 BEGIN {
   eval {
     require XSLoader;
     XSLoader::load(__PACKAGE__, $Math::Prime::Util::GMP::VERSION);
     _GMP_init();
+    _init_random();
     1;
   } or do {
     die $@;
-  }
+  };
 }
 END {
   _GMP_destroy();
 }
 
+
 sub _validate_positive_integer {
   my($n, $min, $max) = @_;
   croak "Parameter must be defined" if !defined $n;
@@ -160,7 +190,7 @@ __END__
 
 =encoding utf8
 
-=for stopwords Möbius Deléglise Bézout gcdext vecsum vecprod moebius totient liouville znorder znprimroot bernfrac bernreal harmfrac harmreal stirling zeta riemannr lucasu lucasv OpenPFGW gmpy2 nonresidue chinese tuplets sqrtmod addmod mulmod powmod divmod superset sqrtint rootint todigits
+=for stopwords Möbius Deléglise Bézout s-gonal gcdext vecsum vecprod moebius totient liouville znorder znprimroot bernfrac bernreal harmfrac harmreal logreal expreal powreal agmreal stirling zeta li ei riemannr lambertw lucasu lucasv OpenPFGW gmpy2 nonresidue chinese tuplets sqrtmod addmod mulmod powmod divmod superset sqrtint rootint logint todigits urandomb urandomr
 
 =head1 NAME
 
@@ -169,7 +199,7 @@ Math::Prime::Util::GMP - Utilities related to prime numbers and factoring, using
 
 =head1 VERSION
 
-Version 0.41
+Version 0.50
 
 
 =head1 SYNOPSIS
@@ -451,7 +481,7 @@ GMP integer, so may be large.  Typically this is not necessary, but
 cryptographic applications may prefer the ability to use this, and it
 allows repeatable test results.
 
-There is no check for duplicate bases.  Input sizes below 65-bits make
+There is no check for duplicate bases.  Input sizes below 65 bits make
 little sense for this function since L<is_prob_prime> is deterministic
 at that size.  For numbers of 65+ bits, the chance of duplicate bases
 is quite small.  The exponentiation approximation for the birthday
@@ -596,7 +626,8 @@ or 1 (probably prime), using the BPSW primality test (extra-strong variant).
 
 This function does the extra-strong BPSW test and nothing more.  That is,
 it will skip all pretests and any extra work that the L</is_prob_prime>
-test may add.
+test may add.  This saves some time if the input has no small factors, such
+as testing results that have been sieved.
 
 
 =head2 is_aks_prime
@@ -611,7 +642,7 @@ The particular method used is theorem 4.1 from Bernstein (2003).  This is
 substantially faster than the original AKS publication, the later version
 with improvements by Lenstra (sometimes called the V6 paper), or the later
 improvements of Voloch and Bornemann.  It is, by a large order, faster than
-any other known implementation as of early 2016.
+any other known implementation as of early 2017.
 
 For theoretical analysis of the primality task, AKS is extremely important.
 In practice, it is essentially useless.  Estimated run time for a 150 digit
@@ -619,6 +650,11 @@ input is over 2 days, making the case that while the algorithmic complexity
 I<growth> is polynomial, the constants are extremely high.  It will take
 years for for numbers that ECPP or APR-CL can prove in seconds.
 
+With the C<verbose> option set to 1, the chosen C<r> and C<s> values are
+printed before the test starts.  With C<verbose> set to 2 or higher, each
+of the C<s> tests results in a C<.> output as the test runs, allowing progress
+to be monitored.
+
 Typically you should use L</is_provable_prime> and let it decide the method.
 
 =head2 is_mersenne_prime
@@ -637,13 +673,12 @@ Takes a positive number C<n> as input and returns one of: 0 (definitely
 composite), 2 (definitely prime), or -1 (test does not indicate anything).
 This implements the Lucas-Lehmer-Riesel test for
 fast deterministic primality testing on numbers of the form C<k * 2^n - 1>.
+If the input is not of this form or if C<k E<gt>= 2^n> then C<-1> will be
+returned as the test does not apply.
 If C<k = 1> then this is a Mersenne number and the Lucas-Lehmer test is used.
-If the number is not of this form, or if C<k E<lt>= 2^n>, then C<-1> will
-be returned as the test does not apply.
-Otherwise, the LLR
-test is performed.  While not as fast as the Lucas-Lehmer test for Mersenne
-numbers, it is almost as fast as a single strong pseudoprime test (i.e.
-Miller-Rabin test) while giving a certain answer.
+Otherwise, the LLR test is performed.  While not as fast as the
+Lucas-Lehmer test for Mersenne numbers, it is almost as fast as a single
+strong pseudoprime test (i.e.  Miller-Rabin test) while giving a certain answer.
 
 =head2 is_proth_prime
 
@@ -651,8 +686,8 @@ Takes a positive number C<n> as input and returns one of: 0 (definitely
 composite), 2 (definitely prime), or -1 (test does not indicate anything).
 This applies Proth's theorem for fast Las Vegas
 primality testing on numbers of the form C<k * 2^n + 1>.
-If the number is not of this form, or if C<k E<lt>= 2^n>, then C<-1> will
-be returned as the test does not apply.
+If the input is not of this form or if C<k E<gt>= 2^n> then C<-1> will be
+returned as the test does not apply.
 Otherwise, a search is performed to find a quadratic nonresidue modulo C<n>.
 If none can be found after a brief search, C<-1> is returned as no conclusion
 can be reached.  Otherwise, Proth's theorem is checked which conclusively
@@ -758,6 +793,26 @@ of the same name in L<Math::Prime::Util>.
 Values above 64-bit are extra-strong BPSW probable primes.
 
 
+=head2 prime_count
+
+Returns the number of primes between 2 and C<n> (single argument)
+or C<lo> and C<hi> given two arguments.  The values are inclusive.
+
+The method is simple sieving followed by primality testing.  This is
+appropriate for small ranges and is useful for very large arguments.
+The L<Math::Prime::Util> module has much more sophisticated methods
+for 64-bit arguments.
+
+
+=head2 prime_count_lower
+
+=head2 prime_count_upper
+
+Returns lower or upper bounds for the prime count of the input C<n>.
+
+Bounds use Dusart 2010, Büthe 2014, Büthe 2015, and Axler 2017.
+
+
 =head2 sieve_primes
 
   my @primes = sieve_primes(2**100, 2**100 + 10000);
@@ -871,7 +926,7 @@ hence the result is a probable prime (using BPSW).
 
 Returns the distances to the previous and next primes of the input C<n>.
 This is slightly more efficient than calling both L</prev_prime> and
-L</next_prime>, and returning the native integer distances can be more
+L</next_prime>, and returning the distances as native integers is more
 efficient with large inputs.
 
 If an optional second argument C<d> is given, and the input C<n> is larger
@@ -885,6 +940,86 @@ Note that with a non-zero second argument, the values returned have not
 undergone a full BPSW test; just sieving and a SPSP-2 test.
 
 
+=head2 random_nbit_prime
+
+  say "random 512-bit prime: ", random_nbit_prime(512);
+
+Returns a randomly selected prime of exactly C<n> bits.
+C<undef> is returned if C<n> is less than C<2>.
+The returned prime has passed the C<is_prob_prime> (extra strong BPSW) test.
+
+=head2 random_strong_prime
+
+  say "random 512-bit strong prime: ", random_strong_prime(512);
+
+Returns a randomly selected strong prime of exactly C<n> bits.
+C<n> must be at least C<256>.
+The returned prime has passed the C<is_prob_prime> (extra strong BPSW) test.
+
+Given the returned prime C<p>, C<p+1>, C<q=p-1>, and C<q-1> will all have
+a large factor.  This makes using factoring methods such as p-1 and p+1 much
+harder.  Gordon's algorithm is used.  The value of using strong primes is
+questionable over proper random primes when the number of bits is
+at least 1024.
+
+=head2 random_ndigit_prime
+
+  say "random 200-digit prime: ", random_ndigit_prime(200);
+
+Returns a randomly selected prime of exactly C<n> digits.
+C<undef> is returned if C<n> is less than C<1>.
+The returned prime has passed the C<is_prob_prime> (extra strong BPSW) test.
+
+=head2 random_prime
+
+  say random_prime(1000, 2000);  # prime between 1000 and 2000 inclusive
+
+Returns a random prime in the interval C<[a,b]> or C<undef> if no prime is
+in the range.
+The returned prime has passed the C<is_prob_prime> (extra strong BPSW) test.
+
+The random prime functions use the internal CSPRNG for randomness.  This
+is currently ISAAC-32 but will likely change to ChaCha20 in a later release.
+
+This corresponds to Mathematica's C<RandomPrime[{min,max}]> function.
+This is a superset of Pari's C<randomprime(n)> function, where our interval
+API is more convenient for cryptographic functions.
+
+=head2 random_maurer_prime
+
+  say "random 512-bit proven prime: ", random_maurer_prime(512);
+
+Returns an n-bit proven prime using Ueli Maurer's FastPrime algorithm (1995).
+This results in uniform random selection of a proven prime, though not every
+n-bit prime can be generated with this algorithm.
+
+C<undef> is returned if C<n> is less than C<2>.
+As a safety check, internally the extra strong BPSW test is additionally
+run on each intermediate and the final result.
+
+=head2 random_shawe_taylor_prime
+
+  say "random 512-bit proven prime: ", random_shawe_taylor_prime(512);
+
+Returns an n-bit proven prime using the Shawe-Taylor algorithm (1986) from
+section C.6 of FIPS 186-4, although using our CSPRNG rather than SHA-256.
+This is a slightly simpler and older (1986) method than Maurer's algorithm.
+It is a bit faster than Maurer's method but has a smaller subset of returned
+primes.
+
+C<undef> is returned if C<n> is less than C<2>.
+As a safety check, internally the extra strong BPSW test is additionally
+run on each intermediate and the final result.
+
+=head2 random_maurer_prime_with_cert
+
+Like L</random_maurer_prime> but also returns a string certificate.
+
+=head2 random_shawe_taylor_prime_with_cert
+
+Like L</random_shawe_taylor_prime> but also returns a string certificate.
+
+
 =head2 lucasu
 
   say "Fibonacci($_) = ", lucasu(1,-1,$_) for 0..100;
@@ -955,6 +1090,12 @@ defined as the product of the integers 1 to C<n> with the special case
 of C<factorial(0) = 1>.  This corresponds to Pari's C<factorial(n)>
 and Mathematica's C<Factorial[n]> functions.
 
+=head2 factorialmod
+
+Given two positive integer arguments C<n> and C<m>, returns C<n! mod m>.
+This is much faster than computing the large C<factorial(n)> followed
+by a mod operation.
+
 =head2 gcd
 
 Given a list of integers, returns the greatest common divisor.  This is
@@ -971,7 +1112,7 @@ and C<u*x + v*y = d>.  This uses the extended Euclidian algorithm to compute
 the values satisfying Bézout's Identity.
 
 This corresponds to Pari's C<gcdext> function, which was renamed from
-C<bezout> out Pari 2.6.  The results will hence match L<Math::Pari/bezout>.
+C<bezout> in Pari 2.6.  The results will hence match L<Math::Pari/bezout>.
 
 =head2 chinese
 
@@ -1022,7 +1163,53 @@ For negative arguments, this matches Mathematica.  Pari does not implement
 the C<n E<lt> 0, k E<lt>= n> extension and instead returns C<0> for this
 case.  GMP's API does not allow negative C<k> but otherwise matches.
 L<Math::BigInt> does not implement any extensions and the results for
-C<n E<lt> 0, k > 0> are undefined.
+C<n E<lt> 0, k E<gt> 0> are undefined.
+
+=head2 logreal
+
+Returns the natural logarithm of the input C<n>.
+An optional second argument indicates the number of significant digits
+(default 40) with the result rounded.
+
+For C<logreal(2)> we use Formula 25 from Gourdon and Sebah (2010).
+For other values we use AGM (Sasaki and Kanada theta method).
+Performance is 100-1000x faster than Math::BigFloat's GMP backend.
+It is 10x slower than Pari/GP 2.10 and MPFR.
+
+Negative inputs are returned as C<-log(-n)>, which matches L<bignum>.
+Pari/GP and Mathematica return C<log(-n) + Pi*i>.
+
+=head2 expreal
+
+Returns C<e^n> for the input C<n>.
+An optional second argument indicates the number of significant digits
+(default 40) with the result rounded.
+
+The implementation computes C<sinh(n)>, then C<e^x> from that.
+
+=head2 powreal
+
+Returns C<n^x> for the inputs C<n> and C<x>.
+An optional third argument indicates the number of significant digits
+(default 40) with the result rounded.
+
+Like L<logreal> and L<expreal>, this is a basic math function that is
+not available from the GMP library but implemented in MPFR.  Since the
+latter is not always available, this can be useful to have.
+
+=head2 agmreal
+
+Returns the Arithmetic-Geometric mean (AGM) of C<a> and C<b>.
+An optional third argument indicates the number of significant digits
+(default 40) with the result rounded.
+
+Examples of use include elementary constants (e.g. C<Pi> and C<e>),
+logs, exponentials, trigonometric functions, elliptic integrals,
+computing pendulum periods, and more.
+
+This corresponds to Pari's C<agm(x,y)> function, limited to positive
+reals (Pari also handles negative, complex, p-adic, and power series
+arguments).
 
 =head2 bernfrac
 
@@ -1038,6 +1225,7 @@ Returns the Bernoulli number C<B_n> for an integer argument C<n>, as a
 string floating point.  An optional second argument indicates the number
 of significant digits to be used, with the result rounded.  The default
 is 40 digits.
+This corresponds to Pari's C<bernreal> function and.
 
 =head2 harmfrac
 
@@ -1045,6 +1233,7 @@ Returns the Harmonic number C<H_n> for an integer argument C<n>, as a
 rational number.  Two values are returned, the numerator and denominator.
 numbers are the sum of reciprocals of the first C<n> natural numbers:
 C<1 + 1/2 + 1/3 + ... + 1/n>.
+This corresponds to Mathematica's C<HarmonicNumber> function.
 
 =head2 harmreal
 
@@ -1078,9 +1267,26 @@ the number of digits past the decimal point (default 40).
 
 The implementation is algorithm 2 of Borwein (1991).  Performance with
 integer inputs is good, but floating point arguments with high precision
-will be much slower than methods using MPFR.  L<Math::Prime::Util> will
+will be slower than methods using MPFR.  L<Math::Prime::Util> will
 try to use L<Math::MPFR> if possible.
 
+=head2 li
+
+Given a positive integer or float C<n>, returns the real Logarithmic Integral
+as a string floating point.  An optional second argument indicates the number
+of significant digits (default 40) with the result rounded.
+
+The implementation uses Ramanjan's series.
+This corresponds to Mathematica's C<Li> function.
+
+=head2 ei
+
+Given a positive integer or float C<n>, returns the real Exponential Integral
+as a string floating point.  An optional second argument indicates the number
+of significant digits (default 40) with the result rounded.
+
+The implementation is simply li(exp(x)).
+
 =head2 riemannr
 
 Given a positive integer or float C<n>, returns the real Riemann R function
@@ -1088,6 +1294,14 @@ as a string floating point.  An optional second argument indicates the number
 of significant digits (default 40) with the result rounded.
 
 The implementation is the standard Gram series.
+This corresponds to Mathematica's C<RiemannR> function.
+
+=head2 lambertw
+
+Given a float C<x>, returns the principal branch of the Lambert W function.
+This solves for C<W> in the equation C<x = W*exp(W)>.  The input must not be
+less than C<-1/e>.  This corresponds to Pari's C<lambertw>
+function and Mathematica's C<ProductLog> / C<LambertW> function.
 
 
 =head2 znorder
@@ -1121,6 +1335,46 @@ Given two non-negative numbers C<a> and C<n>, returns C<1> if C<a> is a
 primitive root modulo C<n>, and C<0> if not.  If C<a> is a primitive root,
 then C<euler_phi(n)> is the smallest C<e> for which C<a^e = 1 mod n>.
 
+=head2 is_semiprime
+
+Given a positive integer C<n>, returns 1 if C<n> is a semiprime, 0 otherwise.
+A semiprime is the product of exactly two primes.
+
+The boolean result is the same as C<scalar(factor(n)) == 2>, but this
+function performs shortcuts that can greatly speed up the operation.
+
+=head2 is_carmichael
+
+Given a positive integer C<n>, returns 1 if C<n> is a Carmichael number,
+0 otherwise.
+These are composites that satisfy C<b^(n-1) ≡ 1 mod n> for all
+C<1 E<lt> b E<lt> n> relatively prime to C<n>.
+Alternately Korselt's theorem says these are composites such that C<n> is
+square-free and C<p-1> divides C<n-1> for all prime divisors C<p> of C<n>.
+
+Inputs greater than 50 digits use a probabilistic test to avoid fully
+factoring the input.
+
+=head2 is_fundamental
+
+Given a positive integer C<n>, returns 1 if C<n> is a fundamental
+discriminant, 0 otherwise.
+
+=head2 is_totient
+
+Given an integer C<n>, returns 1 if there exists an integer C<x> where
+C<euler_phi(x) == n>.
+
+=head2 is_polygonal
+
+Given integers C<x> and C<s>, return 1 if x is an s-gonal number, 0 otherwise.
+C<s> must be greater than 2.
+
+=head2 polygonal_nth
+
+Given integers C<x> and C<s>, return N if C<x> is the C<N-th> s-gonal number,
+0 otherwise.
+
 
 =head2 sigma
 
@@ -1143,7 +1397,8 @@ numbers, perfect numbers, etc.
 
 Takes a positive integer as input and returns the value of Ramanujan's tau
 function.  The result is a signed integer.
-This corresponds to Mathematica's C<RamanujanTau> function.
+This corresponds to Mathematica's C<RamanujanTau> function
+and Pari's C<ramanujantau> function.
 
 
 =head2 valuation
@@ -1156,6 +1411,13 @@ just applied to integers.
 This corresponds to Pari's C<valuation> function.
 C<0> is returned if C<n> or C<k> is one of the values C<-1>, C<0>, or C<1>.
 
+=head2 hammingweight
+
+Given an integer C<n>, returns the binary Hamming weight of C<abs(n)>.  This
+is also called the population count, and is the number of 1s in the binary
+representation.  This corresponds to Pari's C<hammingweight> function for
+C<t_INT> arguments.
+
 =head2 moebius
 
   say "$n is square free" if moebius($n) != 0;
@@ -1248,19 +1510,58 @@ Perl version can produce C<partitions(10_000)> while with
 L<Math::Prime::Util::GMP> it can do C<partitions(220_000)>.  In contrast,
 in about 10 seconds Pari can solve C<numbpart(22_000_000)>.
 
-If you want the enumerated partitions, see L<Integer::Partition>.  It is
-very fast and uses an extremely memory efficient iterator.  It is not,
-however, practical for producing the partition I<number> for values
+If you want the enumerated partitions, see L<Math::Prime::Util/forpart>
+or L<Integer::Partition>.  These are fast and memory efficient iterators,
+but not practical for producing the partition I<number> for values
 over 100 or so.
 
 
+=head2 numtoperm
+
+  @p = numtoperm(10,654321);  # @p=(1,8,2,7,6,5,3,4,9,0)
+
+Given a non-negative integer C<n> and integer C<k>, return the
+rank C<k> lexicographic permutation of C<n> elements.
+C<k> will be interpreted as mod C<n!>.
+
+=head2 permtonum
+
+  $k = permtonum([1,8,2,7,6,5,3,4,9,0]);  # $k = 654321
+
+Given an array reference containing integers from C<0> to C<n>,
+returns the lexicographic permutation rank of the set.  This is
+the inverse of the L</numtoperm> function.  All integers up to
+C<n> must be present.  The result will be between C<0> and C<n!-1>.
+
+
 =head2 Pi
 
 Takes a positive integer argument C<n> and returns the constant Pi with that
 many digits (including the leading 3).  Rounding is performed.
 
-The implementation uses AGM and is only slightly slower than MPFR (which has
-tighter bounds on the intermediate bits and exit conditions).
+The implementation uses either AGM or Ramanujan/Chudnovsky with binary
+splitting, depending on the number of digits.  It is a little over 2x
+faster than MPFR, similar in speed to Pari/GP, but about 1.5x slower than
+Xue's Chudnovsky demo from the GMP web site.
+Specialized programs such as C<y-cruncher> are even faster.
+
+Note there is a non-trivial amount of overhead in turning the result into
+a string, as well as even more if using the L<ntheory> module which
+further converts the result into a Math::BigFloat object.
+
+Called in void context, this just calculates and caches the result.
+
+
+=head2 Euler
+
+Takes a positive integer argument C<n> and returns Euler's constant
+with that many digits.  Rounding is performed.
+
+The implementation is Brent-McMillan algorithm B, just like Pari/GP.
+Performance is about 3x faster than Pari/GP, but 2-10x slower than MPFR
+which uses binary splitting.
+
+Called in void context, this just calculates and caches the result.
 
 
 =head2 exp_mangoldt
@@ -1334,6 +1635,13 @@ and 0 otherwise.  For example, if C<k=2> then this detects perfect squares.
 This corresponds to Pari/GP's C<ispower> function, with the limitations of
 only integer arguments and no third argument may be given to return the root.
 
+=head2 is_square
+
+Given a positive integer C<n>, returns 1 if C<n> is a perfect square,
+0 otherwise.  This is identical to C<is_power(n,2)>.
+
+This corresponds to Pari/GP's C<issquare> function.
+
 =head2 is_prime_power
 
 Given an integer input C<n>, returns C<k> if C<n = p^k> for some prime p,
@@ -1354,6 +1662,12 @@ of C<n>.
 
 This corresponds to Pari/GP's C<sqrtnint> function.
 
+=head2 logint
+
+Given C<n> and C<b>, returns the integer base-C<b> logarithm of C<n>.
+This is the largest integer C<e> such that C<b^e E<lt>= n>.
+
+This corresponds to Pari/GP's C<logint> function.
 
 =head2 factor
 
@@ -1483,8 +1797,8 @@ Factoring will stop when the input is a prime, one factor has been found, or
 the algorithm fails to find a factor with the given smoothness.
 
 This is Pollard's C<p-1> method using a default smoothness of 5M and a
-second stage of C<B2 = 10 * B1>.  It can quickly find a factor C<p> of the input
-C<n> if the number C<p-1> factors into small primes.  For example
+second stage of C<B2 = 10 * B1>.  It can quickly find a factor C<p> of the
+input C<n> if the number C<p-1> factors into small primes.  For example
 C<n = 22095311209999409685885162322219> has the factor C<p = 3916587618943361>,
 where C<p-1 = 2^7 * 5 * 47 * 59 * 3137 * 703499>, so this method will find
 a factor in the first stage if C<B1 E<gt>= 703499> or in the second stage if
@@ -1509,8 +1823,8 @@ succeeded) or the original number (no factor was found).  In either case,
 multiplying @factors yields the original input.  An optional first stage
 smoothness factor (B1) may be given as the second parameter.  This will be
 the smoothness limit B1 for the first stage.
-Factoring will stop when the input is a prime, one factor has been found, or
-the algorithm fails to find a factor with the given smoothness.
+Factoring will stop when the input is a prime, one factor has been found,
+or the algorithm fails to find a factor with the given smoothness.
 
 
 
@@ -1621,6 +1935,128 @@ The base must be at least 2, and is limited to an int.
 Length must be at least zero and is limited to an int.
 
 
+=head2 seed_csprng
+
+Takes a non-negative integer C<nbytes> and a string C<data> as input.
+These are used to seed the internal CSPRNG used for random functions,
+including the random prime functions.  Ideally this is 16-256 bytes of
+good entropy.
+
+Currently the CSPRNG is ISAAC-32, and the maximum number of seed bytes
+used is 1024.  The CSPRNG will likely change to ChaCha20 in a later release.
+
+
+=head2 is_csprng_well_seeded
+
+Returns true if the CSPRNG has been seeded with 16 or more bytes (128 bits).
+There is no measurement of how "good" the input was.
+
+On startup the module will attempt to seed the CSPRNG from
+C</dev/urandom>, so this function will return true if that was
+successful, but false otherwise.
+
+=head2 urandomb
+
+  $n32 = urandomb(32);    # Classic irand32, returns a UV
+  $n   = urandomb(1024);  # Random integer less than 2^1024
+
+Given a number of bits C<b>, returns a random unsigned integer less than C<2^b>.
+The result will be uniformly distributed between C<0> and C<2^b-1> inclusive.
+
+This is similar to the GMP function C<mpz_urandomb>.
+
+=head2 urandomm
+
+  $n = urandomm(100);    # random integer in [0,99]
+  $n = urandomm(1024);   # random integer in [0,1023]
+
+Given a positive integer C<n>, returns a random unsigned integer less than C<n>.
+The results will be uniformly distributed between C<0> and C<n-1> inclusive.
+
+This is similar to the GMP function C<mpz_urandomm>.
+
+=head2 urandomr
+
+  $n  = urandomr(100, 110);        # Random number [100,110]
+  $nb = urandomr(2**24,2**25-1);   # Random 25-bit number
+  $nd = urandomr(10**24,10**25-1); # Random 25-digit number
+
+Given values C<low> and C<high>, returns a uniform random unsigned integer
+in the range C<[low,high]>.  Both inputs must be non-negative.
+If C<low E<gt> high> then function will return C<undef>.
+Note that the range is inclusive, so C<low>, C<high>, and each integer
+between them have an equal probability of appearing.
+
+=head2 irand
+
+  $n32 = irand;     # random 32-bit integer
+
+Returns a random 32-bit integer using the CSPRNG.
+
+Performance is similar to
+L<Math::Random::MTwist/rand> and L<Math::Random::Xorshift>.
+It is somewhat faster than casting system C<rand> to a 32-bit int.
+It is noticeably faster than
+L<Math::Random::ISAAC>,
+L<Math::Random::ISAAC::XS>,
+L<Math::Random::MT>,
+L<Math::Random::MT::Auto>,
+and L<Crypt::PRNG>.
+
+=head2 irand64
+
+  $n64 = irand64;   # random 64-bit integer
+
+Returns a random 64-bit integer using the CSPRNG (on 64-bit Perl).
+
+=head2 drand
+
+  $f = drand;       # random floating point value in [0,1)
+  $r = drand(25);   # random floating point value in [0,25)
+
+Returns a random NV (Perl's native floating point) using the CSPRNG.
+
+The number of bits returned is equal to the mantissa bits of the NV type
+used for the Perl build, with a max of 64.  By default Perl uses doubles
+and the returned values have 53 bits.  If Perl is built with long double
+support and the long doubles have a larger mantissa, then more bits are
+used.
+
+This gives I<substantially> better quality random numbers than the default
+Perl C<rand> function.  Among other things, on modern Perl's, C<rand> uses
+drand48, which gives 32 bits of decent random values and 16 more bits of
+known patterns (e.g. the 48th bit alternates, the 47th has a period
+of 4, etc.).  There are much better choices for standard random number
+generators, such as the Mersenne Twister from L<Math::Random::MTwist>.
+
+Performance is similar to
+L<Math::Random::MTwist/rand> and L<Math::Random::Xorshift>.
+It is 1.5 - 2x slower than core C<rand> (as are the other modules).
+
+=head2 random_bytes
+
+  $str = random_bytes(32);     # 32 random bytes
+
+Given an unsigned number C<n> of bytes, returns a binary string filled with
+random data from the CSPRNG.  Performance for getting 256 byte strings:
+
+    Module/Method                  Rate   Type
+    -------------             ---------   ----------------------
+    Data::Entropy::Algorithms    2027/s   CSPRNG - AES Counter
+    Crypt::Random                6649/s   CSPRNG - /dev/urandom
+    Bytes::Random                9217/s   drand48
+    Bytes::Random::Secure       23043/s   CSPRNG - ISAAC
+    Math::Random::ISAAC::XS     58377/s   CSPRNG - ISAAC
+    rand+pack                   82977/s   drand48
+    Crypt::PRNG                298567/s   CSPRNG - Fortuna
+    Bytes::Random::XS          383354/s   drand48
+    ntheory                    770364/s   CSPRNG - ChaCha20
+    Math::Random::MTwist      1890151/s   Mersenne Twister
+    Math::Prime::Util::GMP    2045715/s   CSPRNG - ISAAC
+
+Each of the CSPRNG modules should be high quality.  There are no known
+flaws in any of ISAAC, AES CTR, ChaCha20, or Fortuna.  The industry
+seems to be standardizing on ChaCha20 (e.g. BSD, Linux, TLS).
 
 
 =head1 SEE ALSO
@@ -1637,7 +2073,7 @@ this module directly is recommended.
 A Perl module with support for the strong Miller-Rabin test, strong
 Lucas-Selfridge test, the BPSW probable prime test, next_prime / prev_prime,
 the AKS primality test, and prime_count.  It uses L<Math::GMPz> to do all
-the calculations, so is faster than pure Perl bignums, but a little slower
+the calculations, so is faster than pure Perl bignums, but a lot slower
 than XS+GMP.  The prime_count function is only usable for very small inputs,
 but the other functions are quite good for big numbers.  Make sure to use
 version 0.05 or newer.
@@ -1663,12 +2099,6 @@ you want to use this.
 Open source APR-CL primality proof implementation.
 Fast primality proving, though without certificates.
 
-=item L<GMP-ECPP|http://sourceforge.net/projects/gmp-ecpp/>.
-An open source ECPP primality proving program.  Slower than this
-module's ECPP for all inputs when the large polynomial set from
-github is used.  Extremely slow once past 300 or so digits.
-There are now better alternatives.
-
 =back
 
 
@@ -1710,7 +2140,8 @@ preprint, Jan 2003.  L<http://cr.yp.to/papers/aks.pdf>
 
 Dana Jacobsen E<lt>dana@acm.orgE<gt>
 
-William Hart wrote the SIMPQS code which is the basis for the QS code.
+Jason Papadopoulos wrote the tinyqs code which is basically unchanged.
+William Hart wrote SIMPQS which is the basis for the QS code.
 
 
 =head1 ACKNOWLEDGEMENTS
@@ -1726,8 +2157,8 @@ and the like happen, and still get decent performance for my purposes.
 
 Ben Buhrow and Jason Papadopoulos deserve special mention for their open
 source factoring tools, which are both readable and fast.  In particular I am
-leveraging their SQUFOF work in the current implementation.  They are a huge
-resource to the community.
+leveraging their work on SQUFOF in the current implementation.
+They are a huge resource to the community.
 
 Jonathan Leto and Bob Kuo, who wrote and distributed the L<Math::Primality>
 module on CPAN.  Their implementation of BPSW provided the motivation I needed
@@ -1735,12 +2166,12 @@ to do it in this module and L<Math::Prime::Util>.  I also used their
 module quite a bit for testing against.
 
 Paul Zimmermann's papers and GMP-ECM code were of great value for my projective
-ECM implementation, as well as the papers by Brent and Montgomery.
+ECM implementation, as well as the many papers by Brent and Montgomery.
 
 
 =head1 COPYRIGHT
 
-Copyright 2011-2016 by Dana Jacobsen E<lt>dana@acm.orgE<gt>
+Copyright 2011-2017 by Dana Jacobsen E<lt>dana@acm.orgE<gt>
 
 This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.
 
